<?php

namespace Drupal\entity_workflow_content\Plugin\WorkflowType;

use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\entity_workflow\WorkflowType\EntityWorkflowTypeBase;

/**
 * @WorkflowType(
 *   id = "entity_workflow_content",
 *   label = @Translation("Content"),
 * )
 */
class EntityWorkflowContent extends EntityWorkflowTypeBase {

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $default_configuration = &drupal_static('EntityWorkflowContent::defaultConfiguration');
    if (isset($default_configuration)) {
      return $default_configuration;
    }

    $supported_entity_types = array_filter($this->entityTypeManager->getDefinitions(), function (EntityTypeInterface $entity_type) {
      return $this->workspaceManager->isEntityTypeSupported($entity_type);
    });

    $default_configuration = [
      'entity_types' => array_fill_keys(array_keys($supported_entity_types), []),
      'default_state' => 'draft',
      'default_transition' => 'new_draft',
      'states' => [
        'draft' => [
          'label' => t('Draft'),
          'weight' => 0,
        ],
        'review' => [
          'label' => t('Ready for Review'),
          'weight' => 30,
        ],
        'approved' => [
          'label' => t('Approved'),
          'weight' => 100,
        ],
      ],
      'transitions' => [
        'new_draft' => [
          'label' => t('New draft'),
          'from' => [],
          'to' => 'draft',
          'weight' => 0,
        ],
        'submit_for_review' => [
          'label' => t('Submit for Review'),
          'from' => ['draft'],
          'to' => 'review',
          'transition_log' => TRUE,
          'action_link' => TRUE,
          'weight' => 10,
        ],
        'approve' => [
          'label' => t('Approve'),
          'from' => ['review'],
          'to' => 'approved',
          'transition_log' => TRUE,
          'action_link' => TRUE,
          'weight' => 100,
        ],
        'back_to_draft' => [
          'label' => t('Back to Draft'),
          'from' => ['review', 'approved'],
          'to' => 'draft',
          'transition_log' => TRUE,
          'action_link' => TRUE,
          'weight' => -100,
        ],
        // Special "skip to the end" transition.
        'approve_immediately' => [
          'label' => t('Approve immediately'),
          'from' => ['draft', 'review'],
          'to' => 'approved',
          'transition_log' => TRUE,
          'transition_log_required' => TRUE,
          'action_link' => TRUE,
          'weight' => 100,
          // This only shows up if 'approve' doesn't.
          'additional_access_callbacks' => ['entity_workflow_content_can_not_approve'],
        ],
      ],
    ] + parent::defaultConfiguration();

    return $default_configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function getPermissions() {
    $perms = [];

    // Add generic permission to transition to any state.
    $perms['use all content transitions'] = [
      'title' => $this->t('%workflow workflow: Use all transitions', [
        '%workflow' => $this->label(),
      ]),
      'description' => $this->t('Allow the user to transition from any state to any state, as defined in the %workflow workflow.', [
        '%workflow' => $this->label(),
      ]),
    ];

    foreach ($this->getTransitions() as $transition) {
      $perms['use content transition ' . $transition->id()] = [
        'title' => $this->t('%workflow workflow: Use %transition transition', [
          '%workflow' => $this->label(),
          '%transition' => $transition->label(),
        ]),
        'description' => $this->t('Allow the user to use the %transition transition, as defined in the %workflow workflow.', [
          '%workflow' => $this->label(),
          '%transition' => $transition->label(),
        ]),
      ];
    }

    return $perms;
  }

}
