<?php

namespace Drupal\entity_workflow\Plugin\Field\FieldType;

use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemBase;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\TypedData\DataDefinition;
use Drupal\Core\TypedData\OptionsProviderInterface;
use Drupal\entity_workflow\Entity\WorkflowTransitionLog;
use Drupal\entity_workflow\EntityWorkflowStateItemInterface;
use Drupal\workflows\Entity\Workflow;
use Drupal\workflows\StateInterface;
use Drupal\workflows\TransitionInterface;
use Drupal\workspaces\WorkspaceInterface;

/**
 * Plugin implementation of the 'entity_workflow_state' field type.
 *
 * @FieldType(
 *   id = "entity_workflow_state",
 *   label = @Translation("Entity Workflow State"),
 *   description = @Translation("Stores the current workflow state of an entity."),
 *   default_widget = "options_select",
 *   default_formatter = "list_default",
 *   list_class = "\Drupal\entity_workflow\EntityWorkflowStateFieldItemList",
 *   cardinality = 1,
 * )
 */
class EntityWorkflowStateItem extends FieldItemBase implements EntityWorkflowStateItemInterface, OptionsProviderInterface {

  /**
   * The original value, used to validate state changes.
   *
   * @var string
   */
  protected $originalValue;

  /**
   * {@inheritdoc}
   */
  public static function propertyDefinitions(FieldStorageDefinitionInterface $field_definition) {
    $properties['value'] = DataDefinition::create('string')
      ->setLabel(t('State'))
      ->setRequired(TRUE);

    return $properties;
  }

  /**
   * {@inheritdoc}
   */
  public static function schema(FieldStorageDefinitionInterface $field_definition) {
    return [
      'columns' => [
        'value' => [
          'type' => 'varchar_ascii',
          'length' => 255,
        ],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultStorageSettings() {
    return [
      'workflow' => '',
      'workflow_callback' => '',
    ] + parent::defaultFieldSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function storageSettingsForm(array &$form, FormStateInterface $form_state, $has_data) {
    $element = [];
    // Allow the workflow to be changed if it's not determined by a callback.
    if (!$this->getSetting('workflow_callback')) {
      $options = [];
      foreach (Workflow::loadMultiple() as $workflow) {
        $options[$workflow->id()] = $workflow->label();
      }

      $element['workflow'] = [
        '#type' => 'select',
        '#title' => $this->t('Workflow'),
        '#options' => $options,
        '#default_value' => $this->getSetting('workflow'),
        '#disabled' => $has_data,
        '#required' => TRUE,
      ];
    }

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function isEmpty() {
    // Note that in this field's case the value will never be empty
    // because of the default returned in applyDefaultValue().
    return $this->value === NULL || $this->value === '';
  }

  /**
   * {@inheritdoc}
   */
  public function applyDefaultValue($notify = TRUE) {
    // @todo Revisit the default value handling.
//    if ($workflow = $this->getWorkflow()) {
//      $initial_state = $workflow->getTypePlugin()->getInitialState();
//      $this->setValue(['value' => $initial_state->id()], $notify);
//    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function setValue($values, $notify = TRUE) {
    if (empty($this->originalValue)) {
      // If no array is given, then the method received just the state value.
      if (isset($values) && !is_array($values)) {
        $values = ['value' => $values];
      }
      // Track the original field value to allow isValid() to validate changes
      // and to react to transitions.
      $this->originalValue = $values['value'];
    }
    parent::setValue($values, $notify);
  }

  /**
   * {@inheritdoc}
   */
  public function isValid() {
    $allowed_states = $this->getAllowedStates($this->originalValue);
    return isset($allowed_states[$this->value]);
  }

  /**
   * {@inheritdoc}
   */
  public function getPossibleValues(AccountInterface $account = NULL) {
    return array_keys($this->getPossibleOptions($account));
  }

  /**
   * {@inheritdoc}
   */
  public function getPossibleOptions(AccountInterface $account = NULL) {
    /** @var \Drupal\workspaces\WorkspaceManagerInterface $workspace_manager */
    $workspace_manager = \Drupal::service('workspaces.manager');

    $workflow = $this->getWorkflow();
    $state_labels = array_map(function (StateInterface $state) use ($workspace_manager) {
      return $state->label($this->getEntity(), $workspace_manager->getActiveWorkspace());
    }, $workflow->getTypePlugin()->getStates());

    return $state_labels;
  }

  /**
   * {@inheritdoc}
   */
  public function getSettableValues(AccountInterface $account = NULL) {
    return array_keys($this->getSettableOptions($account));
  }

  /**
   * {@inheritdoc}
   */
  public function getSettableOptions(AccountInterface $account = NULL) {
    /** @var \Drupal\workspaces\WorkspaceManagerInterface $workspace_manager */
    $workspace_manager = \Drupal::service('workspaces.manager');

    $allowed_states = $this->getAllowedStates($this->value);
    $state_labels = array_map(function (StateInterface $state) use ($workspace_manager) {
      // @todo Temporary safeguard for no active workspace. Needs investigation.
      $active_workspace = $workspace_manager->getActiveWorkspace();
      if (!$active_workspace) {
        $active_workspace = NULL;
      }
      return $state->label($this->getEntity(), $active_workspace);
    }, $allowed_states);

    return $state_labels;
  }

  /**
   * Gets the next allowed states for the given field value.
   *
   * @param string $value
   *   The field value, representing the state ID.
   *
   * @return \Drupal\workflows\StateInterface[]
   *   The allowed states.
   */
  protected function getAllowedStates($value) {
    $allowed_states = [];
    $workflow = $this->getWorkflow();
    foreach ($workflow->getTypePlugin()->getState($value)->getTransitions() as $transition) {
      $state = $transition->to();
      $allowed_states[$state->id()] = $state;
    }

    return $allowed_states;
  }

  /**
   * {@inheritdoc}
   */
  public function getWorkflow() {
    if ($callback = $this->getSetting('workflow_callback')) {
      $workflow_id = call_user_func($callback, $this->getFieldDefinition());
    }
    else {
      $workflow_id = $this->getSetting('workflow');
    }

    return Workflow::load($workflow_id);
  }

  /**
   * {@inheritdoc}
   */
  public function getOriginalStateId() {
    return $this->originalValue;
  }

  /**
   * {@inheritdoc}
   */
  public function getStateId() {
    return $this->value;
  }

  /**
   * {@inheritdoc}
   */
  public function getStateLabel(WorkspaceInterface $workspace = NULL) {
    if (!$this->value) {
      return $this->t('N/A');
    }

    /** @var \Drupal\workspaces\WorkspaceManagerInterface $workspace_manager */
    $workspace_manager = \Drupal::service('workspaces.manager');
    if (!$workspace && $workspace_manager->hasActiveWorkspace()) {
      $workspace = $workspace_manager->getActiveWorkspace();
    }
    $state = $this->getWorkflow()->getTypePlugin()->getState($this->value);
    return $state->label($this->getEntity(), $workspace);
  }

  /**
   * {@inheritdoc}
   */
  public function getTransitions() {
    $state = $this->getWorkflow()->getTypePlugin()->getState($this->value);
    return $state->getTransitions();
  }

  /**
   * {@inheritdoc}
   */
  public function applyTransition(TransitionInterface $transition, $log = '', $context = []) {
    $this->setValue(['value' => $transition->to($this->getEntity())->id()]);
    $this->transition_id = $transition->id();
    $this->log = $log;
    $this->context = $context;
  }

  /**
   * {@inheritdoc}
   */
  public function applyTransitionById($transition_id, $log = '', $context = []) {
    $transition = $this->getWorkflow()->getTypePlugin()->getTransition($transition_id);
    $this->applyTransition($transition, $log, $context);
  }

  /**
   * {@inheritdoc}
   */
  public function preSave() {
    if ($this->value != $this->originalValue) {
      $this->invokeTransitionHook('pre_transition');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function postSave($update) {
    if ($this->value != $this->originalValue) {
      $this->invokeTransitionHook('post_transition');
    }
    $this->originalValue = $this->value;

    // Record state history.
    if (isset($this->transition_id)) {
      \Drupal::entityTypeManager()->getStorage('workflow_transition_log')->writeLogEntry(
        $this->getEntity(),
        $this->getWorkflow()->id(),
        $this->getOriginalStateId(),
        $this->value,
        $this->transition_id,
        $this->log
      );
    }
  }

  /**
   * Invokes a transition hook for the given phase.
   *
   * @param string $phase
   *   The phase: pre_transition OR post_transition.
   */
  protected function invokeTransitionHook($phase) {
    $workflow = $this->getWorkflow();
    if ($workflow->getTypePlugin()->hasTransitionFromStateToState($this->originalValue, $this->value, $this->getEntity())) {
      $transition = $workflow->getTypePlugin()->getTransitionFromStateToState($this->originalValue, $this->value, $this->getEntity());

      // Invoke the hook.
      $entity = $this->getEntity();
      $field_name = $this->getFieldDefinition()->getName();
      $context = $this->context ?? [];
      \Drupal::moduleHandler()->invokeAll('workflows_' . $phase, [$workflow, $transition, $entity, $field_name, $context]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function calculateDependencies(FieldDefinitionInterface $field_definition) {
    $dependencies = [];

    if ($callback = $field_definition->getSetting('workflow_callback')) {
      $workflow_id = call_user_func($callback, $field_definition);
    }
    else {
      $workflow_id = $field_definition->getSetting('workflow');
    }
    if ($workflow_id && ($workflow = Workflow::load($workflow_id))) {
      $dependencies['config'][] = $workflow->getConfigDependencyName();
    }

    return $dependencies;
  }

}
